#  Copyright (c) 2022 AIRBUS and its affiliates.
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this source tree.

from discrete_optimization.rcpsp.rcpsp_model import RCPSPModel
from discrete_optimization.rcpsp.rcpsp_parser import get_data_available, parse_file
from discrete_optimization.rcpsp.robust_rcpsp import (
    MethodBaseRobustification,
    MethodRobustification,
    UncertainRCPSPModel,
    create_poisson_laws_duration,
    create_poisson_laws_resource,
)


def test_uncertain_durations_sm():
    files = get_data_available()
    files = [f for f in files if "j1201_1.sm" in f]  # Single mode RCPSP
    file_path = files[0]

    rcpsp_model: RCPSPModel = parse_file(file_path)

    uncertain_model: UncertainRCPSPModel = UncertainRCPSPModel(
        base_rcpsp_model=rcpsp_model,
        poisson_laws=create_poisson_laws_duration(rcpsp_model),
    )
    worst_model = uncertain_model.create_rcpsp_model(
        MethodRobustification(
            method_base=MethodBaseRobustification.WORST_CASE, percentile=0
        )
    )
    best_model = uncertain_model.create_rcpsp_model(
        MethodRobustification(
            method_base=MethodBaseRobustification.BEST_CASE, percentile=0
        )
    )
    average_model = uncertain_model.create_rcpsp_model(
        MethodRobustification(
            method_base=MethodBaseRobustification.AVERAGE, percentile=0
        )
    )
    sample_model = uncertain_model.create_rcpsp_model(
        MethodRobustification(
            method_base=MethodBaseRobustification.SAMPLE, percentile=0
        )
    )
    percentile_model = uncertain_model.create_rcpsp_model(
        MethodRobustification(
            method_base=MethodBaseRobustification.PERCENTILE, percentile=100
        )
    )


def test_uncertain_resources_sm():
    files = get_data_available()
    files = [f for f in files if "j1201_1.sm" in f]  # Single mode RCPSP
    # files = [f for f in files if 'j1010_5.mm' in f]  # Multi mode RCPSP
    file_path = files[0]
    rcpsp_model: RCPSPModel = parse_file(file_path)
    poisson_laws = create_poisson_laws_resource(rcpsp_model)
    uncertain_model: UncertainRCPSPModel = UncertainRCPSPModel(
        base_rcpsp_model=rcpsp_model, poisson_laws=poisson_laws
    )
    worst_model = uncertain_model.create_rcpsp_model(
        MethodRobustification(
            method_base=MethodBaseRobustification.WORST_CASE, percentile=0
        )
    )
    assert rcpsp_model.mode_details == {
        1: {1: {"duration": 0, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
        2: {1: {"duration": 6, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        3: {1: {"duration": 4, "R1": 0, "R2": 0, "R3": 4, "R4": 0}},
        4: {1: {"duration": 2, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        5: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 6, "R4": 0}},
        6: {1: {"duration": 3, "R1": 1, "R2": 0, "R3": 0, "R4": 0}},
        7: {1: {"duration": 10, "R1": 6, "R2": 0, "R3": 0, "R4": 0}},
        8: {1: {"duration": 2, "R1": 0, "R2": 6, "R3": 0, "R4": 0}},
        9: {1: {"duration": 7, "R1": 0, "R2": 7, "R3": 0, "R4": 0}},
        10: {1: {"duration": 4, "R1": 10, "R2": 0, "R3": 0, "R4": 0}},
        11: {1: {"duration": 6, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        12: {1: {"duration": 4, "R1": 0, "R2": 0, "R3": 9, "R4": 0}},
        13: {1: {"duration": 7, "R1": 0, "R2": 1, "R3": 0, "R4": 0}},
        14: {1: {"duration": 4, "R1": 0, "R2": 0, "R3": 0, "R4": 4}},
        15: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 0, "R4": 1}},
        16: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        17: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 5, "R4": 0}},
        18: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 10, "R4": 0}},
        19: {1: {"duration": 10, "R1": 0, "R2": 8, "R3": 0, "R4": 0}},
        20: {1: {"duration": 5, "R1": 0, "R2": 0, "R3": 0, "R4": 6}},
        21: {1: {"duration": 9, "R1": 5, "R2": 0, "R3": 0, "R4": 0}},
        22: {1: {"duration": 8, "R1": 7, "R2": 0, "R3": 0, "R4": 0}},
        23: {1: {"duration": 3, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        24: {1: {"duration": 3, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        25: {1: {"duration": 7, "R1": 10, "R2": 0, "R3": 0, "R4": 0}},
        26: {1: {"duration": 8, "R1": 0, "R2": 0, "R3": 6, "R4": 0}},
        27: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 9, "R4": 0}},
        28: {1: {"duration": 9, "R1": 4, "R2": 0, "R3": 0, "R4": 0}},
        29: {1: {"duration": 5, "R1": 0, "R2": 5, "R3": 0, "R4": 0}},
        30: {1: {"duration": 2, "R1": 0, "R2": 2, "R3": 0, "R4": 0}},
        31: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 0, "R4": 1}},
        32: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 3, "R4": 0}},
        33: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 3, "R4": 0}},
        34: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 0, "R4": 4}},
        35: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 9, "R4": 0}},
        36: {1: {"duration": 9, "R1": 0, "R2": 5, "R3": 0, "R4": 0}},
        37: {1: {"duration": 10, "R1": 8, "R2": 0, "R3": 0, "R4": 0}},
        38: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 0, "R4": 4}},
        39: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 0, "R4": 6}},
        40: {1: {"duration": 8, "R1": 4, "R2": 0, "R3": 0, "R4": 0}},
        41: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 0, "R4": 1}},
        42: {1: {"duration": 2, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        43: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 0, "R4": 2}},
        44: {1: {"duration": 8, "R1": 0, "R2": 7, "R3": 0, "R4": 0}},
        45: {1: {"duration": 2, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        46: {1: {"duration": 2, "R1": 0, "R2": 6, "R3": 0, "R4": 0}},
        47: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 6, "R4": 0}},
        48: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 0, "R4": 7}},
        49: {1: {"duration": 7, "R1": 0, "R2": 0, "R3": 0, "R4": 3}},
        50: {1: {"duration": 5, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        51: {1: {"duration": 7, "R1": 0, "R2": 0, "R3": 0, "R4": 3}},
        52: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 3, "R4": 0}},
        53: {1: {"duration": 9, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        54: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 0, "R4": 1}},
        55: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 7, "R4": 0}},
        56: {1: {"duration": 10, "R1": 0, "R2": 6, "R3": 0, "R4": 0}},
        57: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 2, "R4": 0}},
        58: {1: {"duration": 8, "R1": 0, "R2": 3, "R3": 0, "R4": 0}},
        59: {1: {"duration": 9, "R1": 4, "R2": 0, "R3": 0, "R4": 0}},
        60: {1: {"duration": 3, "R1": 0, "R2": 8, "R3": 0, "R4": 0}},
        61: {1: {"duration": 4, "R1": 0, "R2": 0, "R3": 0, "R4": 4}},
        62: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        63: {1: {"duration": 7, "R1": 0, "R2": 4, "R3": 0, "R4": 0}},
        64: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 9, "R4": 0}},
        65: {1: {"duration": 2, "R1": 4, "R2": 0, "R3": 0, "R4": 0}},
        66: {1: {"duration": 5, "R1": 4, "R2": 0, "R3": 0, "R4": 0}},
        67: {1: {"duration": 5, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        68: {1: {"duration": 3, "R1": 0, "R2": 3, "R3": 0, "R4": 0}},
        69: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 10, "R4": 0}},
        70: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        71: {1: {"duration": 3, "R1": 0, "R2": 6, "R3": 0, "R4": 0}},
        72: {1: {"duration": 7, "R1": 5, "R2": 0, "R3": 0, "R4": 0}},
        73: {1: {"duration": 6, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        74: {1: {"duration": 1, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        75: {1: {"duration": 7, "R1": 5, "R2": 0, "R3": 0, "R4": 0}},
        76: {1: {"duration": 5, "R1": 0, "R2": 0, "R3": 0, "R4": 6}},
        77: {1: {"duration": 4, "R1": 0, "R2": 6, "R3": 0, "R4": 0}},
        78: {1: {"duration": 5, "R1": 0, "R2": 0, "R3": 2, "R4": 0}},
        79: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 0, "R4": 7}},
        80: {1: {"duration": 10, "R1": 0, "R2": 4, "R3": 0, "R4": 0}},
        81: {1: {"duration": 1, "R1": 7, "R2": 0, "R3": 0, "R4": 0}},
        82: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 4, "R4": 0}},
        83: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        84: {1: {"duration": 8, "R1": 0, "R2": 10, "R3": 0, "R4": 0}},
        85: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        86: {1: {"duration": 9, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        87: {1: {"duration": 2, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        88: {1: {"duration": 7, "R1": 0, "R2": 0, "R3": 10, "R4": 0}},
        89: {1: {"duration": 3, "R1": 0, "R2": 0, "R3": 8, "R4": 0}},
        90: {1: {"duration": 7, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        91: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        92: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 0, "R4": 4}},
        93: {1: {"duration": 5, "R1": 8, "R2": 0, "R3": 0, "R4": 0}},
        94: {1: {"duration": 5, "R1": 4, "R2": 0, "R3": 0, "R4": 0}},
        95: {1: {"duration": 3, "R1": 0, "R2": 0, "R3": 0, "R4": 2}},
        96: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 9, "R4": 0}},
        97: {1: {"duration": 3, "R1": 0, "R2": 9, "R3": 0, "R4": 0}},
        98: {1: {"duration": 5, "R1": 4, "R2": 0, "R3": 0, "R4": 0}},
        99: {1: {"duration": 5, "R1": 0, "R2": 6, "R3": 0, "R4": 0}},
        100: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 0, "R4": 6}},
        101: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 0, "R4": 1}},
        102: {1: {"duration": 5, "R1": 0, "R2": 0, "R3": 0, "R4": 7}},
        103: {1: {"duration": 8, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        104: {1: {"duration": 10, "R1": 5, "R2": 0, "R3": 0, "R4": 0}},
        105: {1: {"duration": 10, "R1": 0, "R2": 3, "R3": 0, "R4": 0}},
        106: {1: {"duration": 6, "R1": 6, "R2": 0, "R3": 0, "R4": 0}},
        107: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 10, "R4": 0}},
        108: {1: {"duration": 2, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        109: {1: {"duration": 2, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        110: {1: {"duration": 10, "R1": 7, "R2": 0, "R3": 0, "R4": 0}},
        111: {1: {"duration": 1, "R1": 8, "R2": 0, "R3": 0, "R4": 0}},
        112: {1: {"duration": 3, "R1": 0, "R2": 0, "R3": 8, "R4": 0}},
        113: {1: {"duration": 4, "R1": 0, "R2": 0, "R3": 0, "R4": 6}},
        114: {1: {"duration": 5, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        115: {1: {"duration": 9, "R1": 5, "R2": 0, "R3": 0, "R4": 0}},
        116: {1: {"duration": 3, "R1": 0, "R2": 4, "R3": 0, "R4": 0}},
        117: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 4, "R4": 0}},
        118: {1: {"duration": 10, "R1": 0, "R2": 9, "R3": 0, "R4": 0}},
        119: {1: {"duration": 3, "R1": 0, "R2": 0, "R3": 0, "R4": 7}},
        120: {1: {"duration": 8, "R1": 0, "R2": 7, "R3": 0, "R4": 0}},
        121: {1: {"duration": 9, "R1": 6, "R2": 0, "R3": 0, "R4": 0}},
        122: {1: {"duration": 0, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
    }
    assert worst_model.mode_details == {
        1: {1: {"duration": 0, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
        2: {1: {"duration": 6, "R1": 10, "R2": 1, "R3": 1, "R4": 1}},
        3: {1: {"duration": 4, "R1": 1, "R2": 1, "R3": 5, "R4": 1}},
        4: {1: {"duration": 2, "R1": 3, "R2": 1, "R3": 1, "R4": 1}},
        5: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 7, "R4": 1}},
        6: {1: {"duration": 3, "R1": 2, "R2": 1, "R3": 1, "R4": 1}},
        7: {1: {"duration": 10, "R1": 7, "R2": 1, "R3": 1, "R4": 1}},
        8: {1: {"duration": 2, "R1": 1, "R2": 7, "R3": 1, "R4": 1}},
        9: {1: {"duration": 7, "R1": 1, "R2": 8, "R3": 1, "R4": 1}},
        10: {1: {"duration": 4, "R1": 11, "R2": 1, "R3": 1, "R4": 1}},
        11: {1: {"duration": 6, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        12: {1: {"duration": 4, "R1": 1, "R2": 1, "R3": 10, "R4": 1}},
        13: {1: {"duration": 7, "R1": 1, "R2": 2, "R3": 1, "R4": 1}},
        14: {1: {"duration": 4, "R1": 1, "R2": 1, "R3": 1, "R4": 5}},
        15: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 1, "R4": 2}},
        16: {1: {"duration": 9, "R1": 1, "R2": 1, "R3": 1, "R4": 6}},
        17: {1: {"duration": 1, "R1": 1, "R2": 1, "R3": 6, "R4": 1}},
        18: {1: {"duration": 9, "R1": 1, "R2": 1, "R3": 11, "R4": 1}},
        19: {1: {"duration": 10, "R1": 1, "R2": 9, "R3": 1, "R4": 1}},
        20: {1: {"duration": 5, "R1": 1, "R2": 1, "R3": 1, "R4": 7}},
        21: {1: {"duration": 9, "R1": 6, "R2": 1, "R3": 1, "R4": 1}},
        22: {1: {"duration": 8, "R1": 8, "R2": 1, "R3": 1, "R4": 1}},
        23: {1: {"duration": 3, "R1": 10, "R2": 1, "R3": 1, "R4": 1}},
        24: {1: {"duration": 3, "R1": 10, "R2": 1, "R3": 1, "R4": 1}},
        25: {1: {"duration": 7, "R1": 11, "R2": 1, "R3": 1, "R4": 1}},
        26: {1: {"duration": 8, "R1": 1, "R2": 1, "R3": 7, "R4": 1}},
        27: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 10, "R4": 1}},
        28: {1: {"duration": 9, "R1": 5, "R2": 1, "R3": 1, "R4": 1}},
        29: {1: {"duration": 5, "R1": 1, "R2": 6, "R3": 1, "R4": 1}},
        30: {1: {"duration": 2, "R1": 1, "R2": 3, "R3": 1, "R4": 1}},
        31: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 1, "R4": 2}},
        32: {1: {"duration": 6, "R1": 1, "R2": 1, "R3": 4, "R4": 1}},
        33: {1: {"duration": 9, "R1": 1, "R2": 1, "R3": 4, "R4": 1}},
        34: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 1, "R4": 5}},
        35: {1: {"duration": 6, "R1": 1, "R2": 1, "R3": 10, "R4": 1}},
        36: {1: {"duration": 9, "R1": 1, "R2": 6, "R3": 1, "R4": 1}},
        37: {1: {"duration": 10, "R1": 9, "R2": 1, "R3": 1, "R4": 1}},
        38: {1: {"duration": 9, "R1": 1, "R2": 1, "R3": 1, "R4": 5}},
        39: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 1, "R4": 7}},
        40: {1: {"duration": 8, "R1": 5, "R2": 1, "R3": 1, "R4": 1}},
        41: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 1, "R4": 2}},
        42: {1: {"duration": 2, "R1": 10, "R2": 1, "R3": 1, "R4": 1}},
        43: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 1, "R4": 3}},
        44: {1: {"duration": 8, "R1": 1, "R2": 8, "R3": 1, "R4": 1}},
        45: {1: {"duration": 2, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        46: {1: {"duration": 2, "R1": 1, "R2": 7, "R3": 1, "R4": 1}},
        47: {1: {"duration": 1, "R1": 1, "R2": 1, "R3": 7, "R4": 1}},
        48: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 1, "R4": 8}},
        49: {1: {"duration": 7, "R1": 1, "R2": 1, "R3": 1, "R4": 4}},
        50: {1: {"duration": 5, "R1": 10, "R2": 1, "R3": 1, "R4": 1}},
        51: {1: {"duration": 7, "R1": 1, "R2": 1, "R3": 1, "R4": 4}},
        52: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 4, "R4": 1}},
        53: {1: {"duration": 9, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        54: {1: {"duration": 9, "R1": 1, "R2": 1, "R3": 1, "R4": 2}},
        55: {1: {"duration": 1, "R1": 1, "R2": 1, "R3": 8, "R4": 1}},
        56: {1: {"duration": 10, "R1": 1, "R2": 7, "R3": 1, "R4": 1}},
        57: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 3, "R4": 1}},
        58: {1: {"duration": 8, "R1": 1, "R2": 4, "R3": 1, "R4": 1}},
        59: {1: {"duration": 9, "R1": 5, "R2": 1, "R3": 1, "R4": 1}},
        60: {1: {"duration": 3, "R1": 1, "R2": 9, "R3": 1, "R4": 1}},
        61: {1: {"duration": 4, "R1": 1, "R2": 1, "R3": 1, "R4": 5}},
        62: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 1, "R4": 6}},
        63: {1: {"duration": 7, "R1": 1, "R2": 5, "R3": 1, "R4": 1}},
        64: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 10, "R4": 1}},
        65: {1: {"duration": 2, "R1": 5, "R2": 1, "R3": 1, "R4": 1}},
        66: {1: {"duration": 5, "R1": 5, "R2": 1, "R3": 1, "R4": 1}},
        67: {1: {"duration": 5, "R1": 3, "R2": 1, "R3": 1, "R4": 1}},
        68: {1: {"duration": 3, "R1": 1, "R2": 4, "R3": 1, "R4": 1}},
        69: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 11, "R4": 1}},
        70: {1: {"duration": 6, "R1": 1, "R2": 1, "R3": 1, "R4": 6}},
        71: {1: {"duration": 3, "R1": 1, "R2": 7, "R3": 1, "R4": 1}},
        72: {1: {"duration": 7, "R1": 6, "R2": 1, "R3": 1, "R4": 1}},
        73: {1: {"duration": 6, "R1": 3, "R2": 1, "R3": 1, "R4": 1}},
        74: {1: {"duration": 1, "R1": 3, "R2": 1, "R3": 1, "R4": 1}},
        75: {1: {"duration": 7, "R1": 6, "R2": 1, "R3": 1, "R4": 1}},
        76: {1: {"duration": 5, "R1": 1, "R2": 1, "R3": 1, "R4": 7}},
        77: {1: {"duration": 4, "R1": 1, "R2": 7, "R3": 1, "R4": 1}},
        78: {1: {"duration": 5, "R1": 1, "R2": 1, "R3": 3, "R4": 1}},
        79: {1: {"duration": 6, "R1": 1, "R2": 1, "R3": 1, "R4": 8}},
        80: {1: {"duration": 10, "R1": 1, "R2": 5, "R3": 1, "R4": 1}},
        81: {1: {"duration": 1, "R1": 8, "R2": 1, "R3": 1, "R4": 1}},
        82: {1: {"duration": 9, "R1": 1, "R2": 1, "R3": 5, "R4": 1}},
        83: {1: {"duration": 1, "R1": 1, "R2": 1, "R3": 1, "R4": 6}},
        84: {1: {"duration": 8, "R1": 1, "R2": 11, "R3": 1, "R4": 1}},
        85: {1: {"duration": 6, "R1": 1, "R2": 1, "R3": 1, "R4": 6}},
        86: {1: {"duration": 9, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        87: {1: {"duration": 2, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        88: {1: {"duration": 7, "R1": 1, "R2": 1, "R3": 11, "R4": 1}},
        89: {1: {"duration": 3, "R1": 1, "R2": 1, "R3": 9, "R4": 1}},
        90: {1: {"duration": 7, "R1": 10, "R2": 1, "R3": 1, "R4": 1}},
        91: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 1, "R4": 6}},
        92: {1: {"duration": 1, "R1": 1, "R2": 1, "R3": 1, "R4": 5}},
        93: {1: {"duration": 5, "R1": 9, "R2": 1, "R3": 1, "R4": 1}},
        94: {1: {"duration": 5, "R1": 5, "R2": 1, "R3": 1, "R4": 1}},
        95: {1: {"duration": 3, "R1": 1, "R2": 1, "R3": 1, "R4": 3}},
        96: {1: {"duration": 1, "R1": 1, "R2": 1, "R3": 10, "R4": 1}},
        97: {1: {"duration": 3, "R1": 1, "R2": 10, "R3": 1, "R4": 1}},
        98: {1: {"duration": 5, "R1": 5, "R2": 1, "R3": 1, "R4": 1}},
        99: {1: {"duration": 5, "R1": 1, "R2": 7, "R3": 1, "R4": 1}},
        100: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 1, "R4": 7}},
        101: {1: {"duration": 6, "R1": 1, "R2": 1, "R3": 1, "R4": 2}},
        102: {1: {"duration": 5, "R1": 1, "R2": 1, "R3": 1, "R4": 8}},
        103: {1: {"duration": 8, "R1": 10, "R2": 1, "R3": 1, "R4": 1}},
        104: {1: {"duration": 10, "R1": 6, "R2": 1, "R3": 1, "R4": 1}},
        105: {1: {"duration": 10, "R1": 1, "R2": 4, "R3": 1, "R4": 1}},
        106: {1: {"duration": 6, "R1": 7, "R2": 1, "R3": 1, "R4": 1}},
        107: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 11, "R4": 1}},
        108: {1: {"duration": 2, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        109: {1: {"duration": 2, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        110: {1: {"duration": 10, "R1": 8, "R2": 1, "R3": 1, "R4": 1}},
        111: {1: {"duration": 1, "R1": 9, "R2": 1, "R3": 1, "R4": 1}},
        112: {1: {"duration": 3, "R1": 1, "R2": 1, "R3": 9, "R4": 1}},
        113: {1: {"duration": 4, "R1": 1, "R2": 1, "R3": 1, "R4": 7}},
        114: {1: {"duration": 5, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        115: {1: {"duration": 9, "R1": 6, "R2": 1, "R3": 1, "R4": 1}},
        116: {1: {"duration": 3, "R1": 1, "R2": 5, "R3": 1, "R4": 1}},
        117: {1: {"duration": 1, "R1": 1, "R2": 1, "R3": 5, "R4": 1}},
        118: {1: {"duration": 10, "R1": 1, "R2": 10, "R3": 1, "R4": 1}},
        119: {1: {"duration": 3, "R1": 1, "R2": 1, "R3": 1, "R4": 8}},
        120: {1: {"duration": 8, "R1": 1, "R2": 8, "R3": 1, "R4": 1}},
        121: {1: {"duration": 9, "R1": 7, "R2": 1, "R3": 1, "R4": 1}},
        122: {1: {"duration": 0, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
    }
    best_model = uncertain_model.create_rcpsp_model(
        MethodRobustification(
            method_base=MethodBaseRobustification.BEST_CASE, percentile=0
        )
    )
    assert best_model.mode_details == {
        1: {1: {"duration": 0, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
        2: {1: {"duration": 6, "R1": 8, "R2": 0, "R3": 0, "R4": 0}},
        3: {1: {"duration": 4, "R1": 0, "R2": 0, "R3": 3, "R4": 0}},
        4: {1: {"duration": 2, "R1": 1, "R2": 0, "R3": 0, "R4": 0}},
        5: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 5, "R4": 0}},
        6: {1: {"duration": 3, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
        7: {1: {"duration": 10, "R1": 5, "R2": 0, "R3": 0, "R4": 0}},
        8: {1: {"duration": 2, "R1": 0, "R2": 5, "R3": 0, "R4": 0}},
        9: {1: {"duration": 7, "R1": 0, "R2": 6, "R3": 0, "R4": 0}},
        10: {1: {"duration": 4, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        11: {1: {"duration": 6, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        12: {1: {"duration": 4, "R1": 0, "R2": 0, "R3": 8, "R4": 0}},
        13: {1: {"duration": 7, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
        14: {1: {"duration": 4, "R1": 0, "R2": 0, "R3": 0, "R4": 3}},
        15: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
        16: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 0, "R4": 4}},
        17: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 4, "R4": 0}},
        18: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 9, "R4": 0}},
        19: {1: {"duration": 10, "R1": 0, "R2": 7, "R3": 0, "R4": 0}},
        20: {1: {"duration": 5, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        21: {1: {"duration": 9, "R1": 4, "R2": 0, "R3": 0, "R4": 0}},
        22: {1: {"duration": 8, "R1": 6, "R2": 0, "R3": 0, "R4": 0}},
        23: {1: {"duration": 3, "R1": 8, "R2": 0, "R3": 0, "R4": 0}},
        24: {1: {"duration": 3, "R1": 8, "R2": 0, "R3": 0, "R4": 0}},
        25: {1: {"duration": 7, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        26: {1: {"duration": 8, "R1": 0, "R2": 0, "R3": 5, "R4": 0}},
        27: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 8, "R4": 0}},
        28: {1: {"duration": 9, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        29: {1: {"duration": 5, "R1": 0, "R2": 4, "R3": 0, "R4": 0}},
        30: {1: {"duration": 2, "R1": 0, "R2": 1, "R3": 0, "R4": 0}},
        31: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
        32: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 2, "R4": 0}},
        33: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 2, "R4": 0}},
        34: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 0, "R4": 3}},
        35: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 8, "R4": 0}},
        36: {1: {"duration": 9, "R1": 0, "R2": 4, "R3": 0, "R4": 0}},
        37: {1: {"duration": 10, "R1": 7, "R2": 0, "R3": 0, "R4": 0}},
        38: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 0, "R4": 3}},
        39: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        40: {1: {"duration": 8, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        41: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
        42: {1: {"duration": 2, "R1": 8, "R2": 0, "R3": 0, "R4": 0}},
        43: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 0, "R4": 1}},
        44: {1: {"duration": 8, "R1": 0, "R2": 6, "R3": 0, "R4": 0}},
        45: {1: {"duration": 2, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        46: {1: {"duration": 2, "R1": 0, "R2": 5, "R3": 0, "R4": 0}},
        47: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 5, "R4": 0}},
        48: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 0, "R4": 6}},
        49: {1: {"duration": 7, "R1": 0, "R2": 0, "R3": 0, "R4": 2}},
        50: {1: {"duration": 5, "R1": 8, "R2": 0, "R3": 0, "R4": 0}},
        51: {1: {"duration": 7, "R1": 0, "R2": 0, "R3": 0, "R4": 2}},
        52: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 2, "R4": 0}},
        53: {1: {"duration": 9, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        54: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
        55: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 6, "R4": 0}},
        56: {1: {"duration": 10, "R1": 0, "R2": 5, "R3": 0, "R4": 0}},
        57: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 1, "R4": 0}},
        58: {1: {"duration": 8, "R1": 0, "R2": 2, "R3": 0, "R4": 0}},
        59: {1: {"duration": 9, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        60: {1: {"duration": 3, "R1": 0, "R2": 7, "R3": 0, "R4": 0}},
        61: {1: {"duration": 4, "R1": 0, "R2": 0, "R3": 0, "R4": 3}},
        62: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 0, "R4": 4}},
        63: {1: {"duration": 7, "R1": 0, "R2": 3, "R3": 0, "R4": 0}},
        64: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 8, "R4": 0}},
        65: {1: {"duration": 2, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        66: {1: {"duration": 5, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        67: {1: {"duration": 5, "R1": 1, "R2": 0, "R3": 0, "R4": 0}},
        68: {1: {"duration": 3, "R1": 0, "R2": 2, "R3": 0, "R4": 0}},
        69: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 9, "R4": 0}},
        70: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 0, "R4": 4}},
        71: {1: {"duration": 3, "R1": 0, "R2": 5, "R3": 0, "R4": 0}},
        72: {1: {"duration": 7, "R1": 4, "R2": 0, "R3": 0, "R4": 0}},
        73: {1: {"duration": 6, "R1": 1, "R2": 0, "R3": 0, "R4": 0}},
        74: {1: {"duration": 1, "R1": 1, "R2": 0, "R3": 0, "R4": 0}},
        75: {1: {"duration": 7, "R1": 4, "R2": 0, "R3": 0, "R4": 0}},
        76: {1: {"duration": 5, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        77: {1: {"duration": 4, "R1": 0, "R2": 5, "R3": 0, "R4": 0}},
        78: {1: {"duration": 5, "R1": 0, "R2": 0, "R3": 1, "R4": 0}},
        79: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 0, "R4": 6}},
        80: {1: {"duration": 10, "R1": 0, "R2": 3, "R3": 0, "R4": 0}},
        81: {1: {"duration": 1, "R1": 6, "R2": 0, "R3": 0, "R4": 0}},
        82: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 3, "R4": 0}},
        83: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 0, "R4": 4}},
        84: {1: {"duration": 8, "R1": 0, "R2": 9, "R3": 0, "R4": 0}},
        85: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 0, "R4": 4}},
        86: {1: {"duration": 9, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        87: {1: {"duration": 2, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        88: {1: {"duration": 7, "R1": 0, "R2": 0, "R3": 9, "R4": 0}},
        89: {1: {"duration": 3, "R1": 0, "R2": 0, "R3": 7, "R4": 0}},
        90: {1: {"duration": 7, "R1": 8, "R2": 0, "R3": 0, "R4": 0}},
        91: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 0, "R4": 4}},
        92: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 0, "R4": 3}},
        93: {1: {"duration": 5, "R1": 7, "R2": 0, "R3": 0, "R4": 0}},
        94: {1: {"duration": 5, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        95: {1: {"duration": 3, "R1": 0, "R2": 0, "R3": 0, "R4": 1}},
        96: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 8, "R4": 0}},
        97: {1: {"duration": 3, "R1": 0, "R2": 8, "R3": 0, "R4": 0}},
        98: {1: {"duration": 5, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        99: {1: {"duration": 5, "R1": 0, "R2": 5, "R3": 0, "R4": 0}},
        100: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        101: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
        102: {1: {"duration": 5, "R1": 0, "R2": 0, "R3": 0, "R4": 6}},
        103: {1: {"duration": 8, "R1": 8, "R2": 0, "R3": 0, "R4": 0}},
        104: {1: {"duration": 10, "R1": 4, "R2": 0, "R3": 0, "R4": 0}},
        105: {1: {"duration": 10, "R1": 0, "R2": 2, "R3": 0, "R4": 0}},
        106: {1: {"duration": 6, "R1": 5, "R2": 0, "R3": 0, "R4": 0}},
        107: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 9, "R4": 0}},
        108: {1: {"duration": 2, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        109: {1: {"duration": 2, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        110: {1: {"duration": 10, "R1": 6, "R2": 0, "R3": 0, "R4": 0}},
        111: {1: {"duration": 1, "R1": 7, "R2": 0, "R3": 0, "R4": 0}},
        112: {1: {"duration": 3, "R1": 0, "R2": 0, "R3": 7, "R4": 0}},
        113: {1: {"duration": 4, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        114: {1: {"duration": 5, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        115: {1: {"duration": 9, "R1": 4, "R2": 0, "R3": 0, "R4": 0}},
        116: {1: {"duration": 3, "R1": 0, "R2": 3, "R3": 0, "R4": 0}},
        117: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 3, "R4": 0}},
        118: {1: {"duration": 10, "R1": 0, "R2": 8, "R3": 0, "R4": 0}},
        119: {1: {"duration": 3, "R1": 0, "R2": 0, "R3": 0, "R4": 6}},
        120: {1: {"duration": 8, "R1": 0, "R2": 6, "R3": 0, "R4": 0}},
        121: {1: {"duration": 9, "R1": 5, "R2": 0, "R3": 0, "R4": 0}},
        122: {1: {"duration": 0, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
    }

    average_model = uncertain_model.create_rcpsp_model(
        MethodRobustification(
            method_base=MethodBaseRobustification.AVERAGE, percentile=0
        )
    )
    assert average_model.mode_details == {
        1: {1: {"duration": 0, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
        2: {1: {"duration": 6, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        3: {1: {"duration": 4, "R1": 0, "R2": 0, "R3": 3, "R4": 0}},
        4: {1: {"duration": 2, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        5: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 6, "R4": 0}},
        6: {1: {"duration": 3, "R1": 1, "R2": 0, "R3": 0, "R4": 0}},
        7: {1: {"duration": 10, "R1": 6, "R2": 0, "R3": 0, "R4": 0}},
        8: {1: {"duration": 2, "R1": 0, "R2": 6, "R3": 0, "R4": 0}},
        9: {1: {"duration": 7, "R1": 0, "R2": 7, "R3": 0, "R4": 0}},
        10: {1: {"duration": 4, "R1": 10, "R2": 0, "R3": 0, "R4": 0}},
        11: {1: {"duration": 6, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        12: {1: {"duration": 4, "R1": 0, "R2": 0, "R3": 9, "R4": 0}},
        13: {1: {"duration": 7, "R1": 0, "R2": 1, "R3": 0, "R4": 0}},
        14: {1: {"duration": 4, "R1": 0, "R2": 0, "R3": 0, "R4": 3}},
        15: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 0, "R4": 1}},
        16: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        17: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 5, "R4": 0}},
        18: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 10, "R4": 0}},
        19: {1: {"duration": 10, "R1": 0, "R2": 8, "R3": 0, "R4": 0}},
        20: {1: {"duration": 5, "R1": 0, "R2": 0, "R3": 0, "R4": 6}},
        21: {1: {"duration": 9, "R1": 5, "R2": 0, "R3": 0, "R4": 0}},
        22: {1: {"duration": 8, "R1": 7, "R2": 0, "R3": 0, "R4": 0}},
        23: {1: {"duration": 3, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        24: {1: {"duration": 3, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        25: {1: {"duration": 7, "R1": 10, "R2": 0, "R3": 0, "R4": 0}},
        26: {1: {"duration": 8, "R1": 0, "R2": 0, "R3": 6, "R4": 0}},
        27: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 9, "R4": 0}},
        28: {1: {"duration": 9, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        29: {1: {"duration": 5, "R1": 0, "R2": 5, "R3": 0, "R4": 0}},
        30: {1: {"duration": 2, "R1": 0, "R2": 2, "R3": 0, "R4": 0}},
        31: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 0, "R4": 1}},
        32: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 3, "R4": 0}},
        33: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 3, "R4": 0}},
        34: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 0, "R4": 3}},
        35: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 9, "R4": 0}},
        36: {1: {"duration": 9, "R1": 0, "R2": 5, "R3": 0, "R4": 0}},
        37: {1: {"duration": 10, "R1": 8, "R2": 0, "R3": 0, "R4": 0}},
        38: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 0, "R4": 3}},
        39: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 0, "R4": 6}},
        40: {1: {"duration": 8, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        41: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 0, "R4": 1}},
        42: {1: {"duration": 2, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        43: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 0, "R4": 2}},
        44: {1: {"duration": 8, "R1": 0, "R2": 7, "R3": 0, "R4": 0}},
        45: {1: {"duration": 2, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        46: {1: {"duration": 2, "R1": 0, "R2": 6, "R3": 0, "R4": 0}},
        47: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 6, "R4": 0}},
        48: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 0, "R4": 7}},
        49: {1: {"duration": 7, "R1": 0, "R2": 0, "R3": 0, "R4": 3}},
        50: {1: {"duration": 5, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        51: {1: {"duration": 7, "R1": 0, "R2": 0, "R3": 0, "R4": 3}},
        52: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 3, "R4": 0}},
        53: {1: {"duration": 9, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        54: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 0, "R4": 1}},
        55: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 7, "R4": 0}},
        56: {1: {"duration": 10, "R1": 0, "R2": 6, "R3": 0, "R4": 0}},
        57: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 2, "R4": 0}},
        58: {1: {"duration": 8, "R1": 0, "R2": 3, "R3": 0, "R4": 0}},
        59: {1: {"duration": 9, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        60: {1: {"duration": 3, "R1": 0, "R2": 8, "R3": 0, "R4": 0}},
        61: {1: {"duration": 4, "R1": 0, "R2": 0, "R3": 0, "R4": 3}},
        62: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        63: {1: {"duration": 7, "R1": 0, "R2": 3, "R3": 0, "R4": 0}},
        64: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 9, "R4": 0}},
        65: {1: {"duration": 2, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        66: {1: {"duration": 5, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        67: {1: {"duration": 5, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        68: {1: {"duration": 3, "R1": 0, "R2": 3, "R3": 0, "R4": 0}},
        69: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 10, "R4": 0}},
        70: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        71: {1: {"duration": 3, "R1": 0, "R2": 6, "R3": 0, "R4": 0}},
        72: {1: {"duration": 7, "R1": 5, "R2": 0, "R3": 0, "R4": 0}},
        73: {1: {"duration": 6, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        74: {1: {"duration": 1, "R1": 2, "R2": 0, "R3": 0, "R4": 0}},
        75: {1: {"duration": 7, "R1": 5, "R2": 0, "R3": 0, "R4": 0}},
        76: {1: {"duration": 5, "R1": 0, "R2": 0, "R3": 0, "R4": 6}},
        77: {1: {"duration": 4, "R1": 0, "R2": 6, "R3": 0, "R4": 0}},
        78: {1: {"duration": 5, "R1": 0, "R2": 0, "R3": 2, "R4": 0}},
        79: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 0, "R4": 7}},
        80: {1: {"duration": 10, "R1": 0, "R2": 3, "R3": 0, "R4": 0}},
        81: {1: {"duration": 1, "R1": 7, "R2": 0, "R3": 0, "R4": 0}},
        82: {1: {"duration": 9, "R1": 0, "R2": 0, "R3": 3, "R4": 0}},
        83: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        84: {1: {"duration": 8, "R1": 0, "R2": 10, "R3": 0, "R4": 0}},
        85: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        86: {1: {"duration": 9, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        87: {1: {"duration": 2, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        88: {1: {"duration": 7, "R1": 0, "R2": 0, "R3": 10, "R4": 0}},
        89: {1: {"duration": 3, "R1": 0, "R2": 0, "R3": 8, "R4": 0}},
        90: {1: {"duration": 7, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        91: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 0, "R4": 5}},
        92: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 0, "R4": 3}},
        93: {1: {"duration": 5, "R1": 8, "R2": 0, "R3": 0, "R4": 0}},
        94: {1: {"duration": 5, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        95: {1: {"duration": 3, "R1": 0, "R2": 0, "R3": 0, "R4": 2}},
        96: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 9, "R4": 0}},
        97: {1: {"duration": 3, "R1": 0, "R2": 9, "R3": 0, "R4": 0}},
        98: {1: {"duration": 5, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        99: {1: {"duration": 5, "R1": 0, "R2": 6, "R3": 0, "R4": 0}},
        100: {1: {"duration": 10, "R1": 0, "R2": 0, "R3": 0, "R4": 6}},
        101: {1: {"duration": 6, "R1": 0, "R2": 0, "R3": 0, "R4": 1}},
        102: {1: {"duration": 5, "R1": 0, "R2": 0, "R3": 0, "R4": 7}},
        103: {1: {"duration": 8, "R1": 9, "R2": 0, "R3": 0, "R4": 0}},
        104: {1: {"duration": 10, "R1": 5, "R2": 0, "R3": 0, "R4": 0}},
        105: {1: {"duration": 10, "R1": 0, "R2": 3, "R3": 0, "R4": 0}},
        106: {1: {"duration": 6, "R1": 6, "R2": 0, "R3": 0, "R4": 0}},
        107: {1: {"duration": 2, "R1": 0, "R2": 0, "R3": 10, "R4": 0}},
        108: {1: {"duration": 2, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        109: {1: {"duration": 2, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        110: {1: {"duration": 10, "R1": 7, "R2": 0, "R3": 0, "R4": 0}},
        111: {1: {"duration": 1, "R1": 8, "R2": 0, "R3": 0, "R4": 0}},
        112: {1: {"duration": 3, "R1": 0, "R2": 0, "R3": 8, "R4": 0}},
        113: {1: {"duration": 4, "R1": 0, "R2": 0, "R3": 0, "R4": 6}},
        114: {1: {"duration": 5, "R1": 3, "R2": 0, "R3": 0, "R4": 0}},
        115: {1: {"duration": 9, "R1": 5, "R2": 0, "R3": 0, "R4": 0}},
        116: {1: {"duration": 3, "R1": 0, "R2": 3, "R3": 0, "R4": 0}},
        117: {1: {"duration": 1, "R1": 0, "R2": 0, "R3": 3, "R4": 0}},
        118: {1: {"duration": 10, "R1": 0, "R2": 9, "R3": 0, "R4": 0}},
        119: {1: {"duration": 3, "R1": 0, "R2": 0, "R3": 0, "R4": 7}},
        120: {1: {"duration": 8, "R1": 0, "R2": 7, "R3": 0, "R4": 0}},
        121: {1: {"duration": 9, "R1": 6, "R2": 0, "R3": 0, "R4": 0}},
        122: {1: {"duration": 0, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
    }

    sample_model = uncertain_model.create_rcpsp_model(
        MethodRobustification(
            method_base=MethodBaseRobustification.SAMPLE, percentile=0
        )
    )

    percentile_model = uncertain_model.create_rcpsp_model(
        MethodRobustification(
            method_base=MethodBaseRobustification.PERCENTILE, percentile=100
        )
    )
    assert percentile_model.mode_details == {
        1: {1: {"duration": 0, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
        2: {1: {"duration": 6, "R1": 10, "R2": 1, "R3": 1, "R4": 1}},
        3: {1: {"duration": 4, "R1": 1, "R2": 1, "R3": 5, "R4": 1}},
        4: {1: {"duration": 2, "R1": 3, "R2": 1, "R3": 1, "R4": 1}},
        5: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 7, "R4": 1}},
        6: {1: {"duration": 3, "R1": 2, "R2": 1, "R3": 1, "R4": 1}},
        7: {1: {"duration": 10, "R1": 7, "R2": 1, "R3": 1, "R4": 1}},
        8: {1: {"duration": 2, "R1": 1, "R2": 7, "R3": 1, "R4": 1}},
        9: {1: {"duration": 7, "R1": 1, "R2": 8, "R3": 1, "R4": 1}},
        10: {1: {"duration": 4, "R1": 11, "R2": 1, "R3": 1, "R4": 1}},
        11: {1: {"duration": 6, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        12: {1: {"duration": 4, "R1": 1, "R2": 1, "R3": 10, "R4": 1}},
        13: {1: {"duration": 7, "R1": 1, "R2": 2, "R3": 1, "R4": 1}},
        14: {1: {"duration": 4, "R1": 1, "R2": 1, "R3": 1, "R4": 5}},
        15: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 1, "R4": 2}},
        16: {1: {"duration": 9, "R1": 1, "R2": 1, "R3": 1, "R4": 6}},
        17: {1: {"duration": 1, "R1": 1, "R2": 1, "R3": 6, "R4": 1}},
        18: {1: {"duration": 9, "R1": 1, "R2": 1, "R3": 11, "R4": 1}},
        19: {1: {"duration": 10, "R1": 1, "R2": 9, "R3": 1, "R4": 1}},
        20: {1: {"duration": 5, "R1": 1, "R2": 1, "R3": 1, "R4": 7}},
        21: {1: {"duration": 9, "R1": 6, "R2": 1, "R3": 1, "R4": 1}},
        22: {1: {"duration": 8, "R1": 8, "R2": 1, "R3": 1, "R4": 1}},
        23: {1: {"duration": 3, "R1": 10, "R2": 1, "R3": 1, "R4": 1}},
        24: {1: {"duration": 3, "R1": 10, "R2": 1, "R3": 1, "R4": 1}},
        25: {1: {"duration": 7, "R1": 11, "R2": 1, "R3": 1, "R4": 1}},
        26: {1: {"duration": 8, "R1": 1, "R2": 1, "R3": 7, "R4": 1}},
        27: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 10, "R4": 1}},
        28: {1: {"duration": 9, "R1": 5, "R2": 1, "R3": 1, "R4": 1}},
        29: {1: {"duration": 5, "R1": 1, "R2": 6, "R3": 1, "R4": 1}},
        30: {1: {"duration": 2, "R1": 1, "R2": 3, "R3": 1, "R4": 1}},
        31: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 1, "R4": 2}},
        32: {1: {"duration": 6, "R1": 1, "R2": 1, "R3": 4, "R4": 1}},
        33: {1: {"duration": 9, "R1": 1, "R2": 1, "R3": 4, "R4": 1}},
        34: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 1, "R4": 5}},
        35: {1: {"duration": 6, "R1": 1, "R2": 1, "R3": 10, "R4": 1}},
        36: {1: {"duration": 9, "R1": 1, "R2": 6, "R3": 1, "R4": 1}},
        37: {1: {"duration": 10, "R1": 9, "R2": 1, "R3": 1, "R4": 1}},
        38: {1: {"duration": 9, "R1": 1, "R2": 1, "R3": 1, "R4": 5}},
        39: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 1, "R4": 7}},
        40: {1: {"duration": 8, "R1": 5, "R2": 1, "R3": 1, "R4": 1}},
        41: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 1, "R4": 2}},
        42: {1: {"duration": 2, "R1": 10, "R2": 1, "R3": 1, "R4": 1}},
        43: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 1, "R4": 3}},
        44: {1: {"duration": 8, "R1": 1, "R2": 8, "R3": 1, "R4": 1}},
        45: {1: {"duration": 2, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        46: {1: {"duration": 2, "R1": 1, "R2": 7, "R3": 1, "R4": 1}},
        47: {1: {"duration": 1, "R1": 1, "R2": 1, "R3": 7, "R4": 1}},
        48: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 1, "R4": 8}},
        49: {1: {"duration": 7, "R1": 1, "R2": 1, "R3": 1, "R4": 4}},
        50: {1: {"duration": 5, "R1": 10, "R2": 1, "R3": 1, "R4": 1}},
        51: {1: {"duration": 7, "R1": 1, "R2": 1, "R3": 1, "R4": 4}},
        52: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 4, "R4": 1}},
        53: {1: {"duration": 9, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        54: {1: {"duration": 9, "R1": 1, "R2": 1, "R3": 1, "R4": 2}},
        55: {1: {"duration": 1, "R1": 1, "R2": 1, "R3": 8, "R4": 1}},
        56: {1: {"duration": 10, "R1": 1, "R2": 7, "R3": 1, "R4": 1}},
        57: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 3, "R4": 1}},
        58: {1: {"duration": 8, "R1": 1, "R2": 4, "R3": 1, "R4": 1}},
        59: {1: {"duration": 9, "R1": 5, "R2": 1, "R3": 1, "R4": 1}},
        60: {1: {"duration": 3, "R1": 1, "R2": 9, "R3": 1, "R4": 1}},
        61: {1: {"duration": 4, "R1": 1, "R2": 1, "R3": 1, "R4": 5}},
        62: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 1, "R4": 6}},
        63: {1: {"duration": 7, "R1": 1, "R2": 5, "R3": 1, "R4": 1}},
        64: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 10, "R4": 1}},
        65: {1: {"duration": 2, "R1": 5, "R2": 1, "R3": 1, "R4": 1}},
        66: {1: {"duration": 5, "R1": 5, "R2": 1, "R3": 1, "R4": 1}},
        67: {1: {"duration": 5, "R1": 3, "R2": 1, "R3": 1, "R4": 1}},
        68: {1: {"duration": 3, "R1": 1, "R2": 4, "R3": 1, "R4": 1}},
        69: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 11, "R4": 1}},
        70: {1: {"duration": 6, "R1": 1, "R2": 1, "R3": 1, "R4": 6}},
        71: {1: {"duration": 3, "R1": 1, "R2": 7, "R3": 1, "R4": 1}},
        72: {1: {"duration": 7, "R1": 6, "R2": 1, "R3": 1, "R4": 1}},
        73: {1: {"duration": 6, "R1": 3, "R2": 1, "R3": 1, "R4": 1}},
        74: {1: {"duration": 1, "R1": 3, "R2": 1, "R3": 1, "R4": 1}},
        75: {1: {"duration": 7, "R1": 6, "R2": 1, "R3": 1, "R4": 1}},
        76: {1: {"duration": 5, "R1": 1, "R2": 1, "R3": 1, "R4": 7}},
        77: {1: {"duration": 4, "R1": 1, "R2": 7, "R3": 1, "R4": 1}},
        78: {1: {"duration": 5, "R1": 1, "R2": 1, "R3": 3, "R4": 1}},
        79: {1: {"duration": 6, "R1": 1, "R2": 1, "R3": 1, "R4": 8}},
        80: {1: {"duration": 10, "R1": 1, "R2": 5, "R3": 1, "R4": 1}},
        81: {1: {"duration": 1, "R1": 8, "R2": 1, "R3": 1, "R4": 1}},
        82: {1: {"duration": 9, "R1": 1, "R2": 1, "R3": 5, "R4": 1}},
        83: {1: {"duration": 1, "R1": 1, "R2": 1, "R3": 1, "R4": 6}},
        84: {1: {"duration": 8, "R1": 1, "R2": 11, "R3": 1, "R4": 1}},
        85: {1: {"duration": 6, "R1": 1, "R2": 1, "R3": 1, "R4": 6}},
        86: {1: {"duration": 9, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        87: {1: {"duration": 2, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        88: {1: {"duration": 7, "R1": 1, "R2": 1, "R3": 11, "R4": 1}},
        89: {1: {"duration": 3, "R1": 1, "R2": 1, "R3": 9, "R4": 1}},
        90: {1: {"duration": 7, "R1": 10, "R2": 1, "R3": 1, "R4": 1}},
        91: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 1, "R4": 6}},
        92: {1: {"duration": 1, "R1": 1, "R2": 1, "R3": 1, "R4": 5}},
        93: {1: {"duration": 5, "R1": 9, "R2": 1, "R3": 1, "R4": 1}},
        94: {1: {"duration": 5, "R1": 5, "R2": 1, "R3": 1, "R4": 1}},
        95: {1: {"duration": 3, "R1": 1, "R2": 1, "R3": 1, "R4": 3}},
        96: {1: {"duration": 1, "R1": 1, "R2": 1, "R3": 10, "R4": 1}},
        97: {1: {"duration": 3, "R1": 1, "R2": 10, "R3": 1, "R4": 1}},
        98: {1: {"duration": 5, "R1": 5, "R2": 1, "R3": 1, "R4": 1}},
        99: {1: {"duration": 5, "R1": 1, "R2": 7, "R3": 1, "R4": 1}},
        100: {1: {"duration": 10, "R1": 1, "R2": 1, "R3": 1, "R4": 7}},
        101: {1: {"duration": 6, "R1": 1, "R2": 1, "R3": 1, "R4": 2}},
        102: {1: {"duration": 5, "R1": 1, "R2": 1, "R3": 1, "R4": 8}},
        103: {1: {"duration": 8, "R1": 10, "R2": 1, "R3": 1, "R4": 1}},
        104: {1: {"duration": 10, "R1": 6, "R2": 1, "R3": 1, "R4": 1}},
        105: {1: {"duration": 10, "R1": 1, "R2": 4, "R3": 1, "R4": 1}},
        106: {1: {"duration": 6, "R1": 7, "R2": 1, "R3": 1, "R4": 1}},
        107: {1: {"duration": 2, "R1": 1, "R2": 1, "R3": 11, "R4": 1}},
        108: {1: {"duration": 2, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        109: {1: {"duration": 2, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        110: {1: {"duration": 10, "R1": 8, "R2": 1, "R3": 1, "R4": 1}},
        111: {1: {"duration": 1, "R1": 9, "R2": 1, "R3": 1, "R4": 1}},
        112: {1: {"duration": 3, "R1": 1, "R2": 1, "R3": 9, "R4": 1}},
        113: {1: {"duration": 4, "R1": 1, "R2": 1, "R3": 1, "R4": 7}},
        114: {1: {"duration": 5, "R1": 4, "R2": 1, "R3": 1, "R4": 1}},
        115: {1: {"duration": 9, "R1": 6, "R2": 1, "R3": 1, "R4": 1}},
        116: {1: {"duration": 3, "R1": 1, "R2": 5, "R3": 1, "R4": 1}},
        117: {1: {"duration": 1, "R1": 1, "R2": 1, "R3": 5, "R4": 1}},
        118: {1: {"duration": 10, "R1": 1, "R2": 10, "R3": 1, "R4": 1}},
        119: {1: {"duration": 3, "R1": 1, "R2": 1, "R3": 1, "R4": 8}},
        120: {1: {"duration": 8, "R1": 1, "R2": 8, "R3": 1, "R4": 1}},
        121: {1: {"duration": 9, "R1": 7, "R2": 1, "R3": 1, "R4": 1}},
        122: {1: {"duration": 0, "R1": 0, "R2": 0, "R3": 0, "R4": 0}},
    }


if __name__ == "__main__":
    test_uncertain_resources_sm()
